import numpy as np

#-------------------------- Fords's algorithm ----------------------------------#


class Node:
    """
        Create node class.
        Instances: ID, predecessing node and value.

        Method: backtrack from final node through its predecessor to initial node
        to describe longest path.
    """

    def __init__(self, id):
        self.id = id
        self.pred = self
        self.val = -np.inf

    def backtrack(self):
        # call predecessors recursively until start node.
        if (self.pred is not self):
            s = self.pred.backtrack() + str('-> %s ' % str(self.id))
            return s
        else:
            return '0'

    def __repr__(self):
        return '(%s,%s,%s)' % (str(self.id), str(self.pred.id), str(self.val))


class Arc:
    """
        Create arc class. Each arc is moving from one arc to another. The cost of the
        arc correspond to the contract revenue of the node its moving to.

        Instances: From Node, To Node and Cost for To Node.
    """

    def __init__(self, nodeFrom, nodeTo, cost):
        # instance is a Node object with attributes (id,pred,val)
        self.nodeFrom = nodeFrom
        self.nodeTo = nodeTo
        self.cost = cost

    def __repr__(self):
        return 'Arc(%s,%s), Cost: %s' % (self.nodeFrom.id, self.nodeTo.id, str(self.cost))


class Network:
    """
        Input: set of node objects and arc objects

        Instances: nodes, arcs, start node, end node, set of solved nodes and set of unsolved
        nodes.

        Methods: step, maxNode, bestNode, arcsFromNode, solve

        step() is the iterative method, which is called by solve() whilst there still are
        unsolved nodes. Each iteration solves one node.

        Ford's algorithm:
        STEP 0 - divide set of nodes into subsets solved (D = []) and unsolved nodes (A).

        STEP 1 - select node n_i which has highest value (first iteration this is node 0).
        Find all existing arcs that are going from this node, (i,j) in B.

        STEP 2 - for all arcs (i,j) in B:
                    if value of node i + arc(i,j) cost is higher than node j value:
                        then a longer path has been found from start, to node j, through node i
                        update node j value = n_i + cost(i,j)                   (I)
                        update n_j predecessor = n_i                            (II)

                            Extra Ford condition (distinguish from Dijkstra)
                            if node j is in solved set D:
                                move node j back to unsolved set A              (III)
        STEP 3 - move n_i to solved set D
        STEP 4 - while subset A is non-empty - go to STEP 1

    """

    def __init__(self, nodes, arcs):
        # STEP 0
        self.nodes = nodes
        self.arcs = arcs
        self.startNode = nodes[0]
        self.endNode = nodes[-1]
        self.solved = []
        self.unsolved = self.nodes

    def step(self):

        # STEP 1
        mx = self.maxNode()
        #print('\nCurrent node: ' + str(mx.id))
        arcsfrom = self.arcsFromNode(mx)
        # STEP 2
        for a in arcsfrom:  # a is arc (from mx,j,cost)
            if a.nodeTo.val < mx.val + a.cost:  # if new cost higher than previous
                a.nodeTo.val = mx.val + a.cost  # (I)
                a.nodeTo.pred = mx  # (II)
                #print('Node ' + str(a.nodeTo.id) + ' updated: ' + str(a.nodeTo))

                # Ford's algorithm added condition: Lundgren p 192, (*)
                if a.nodeTo in self.solved:
                    self.solved.remove(a.nodeTo)  # (III)
                    self.unsolved.append(a.nodeTo)
        # STEP 3
        self.solved.append(mx)
        self.unsolved.remove(mx)
        #print('Solved: %s' % self.solved)
        #print('Unsolved: %s' % self.unsolved)

    def maxNode(self):
        """
            Description:
            Check list of unsolved nodes (id,pred,val), pick node with highest value

            Output: position of node with highest cost.
        """
        i = np.argmax([n.val for n in self.unsolved])
        return self.unsolved[i]

    def arcsFromNode(self, node):
        """
            Description: Find all arcs going from current node. Check list of
            arcs(from,to,cost)

            Input: node
            Output: all arcs going from node

        """
        return [x for x in self.arcs if x.nodeFrom is node]

    def solve(self, rig):
        self.startNode.val = 0
        # STEP 4
        while len(self.unsolved) > 0:
            self.step()
            # print(str(self.endNode.val))
        s = self.endNode.backtrack()
        ###print('\nOffshore unit ' + str(rig.id) + ': ' + s + ' with total profit ' + str(self.endNode.val))

    def result(self):
        """
            Return profit and contracts serviced for evaluated rig.
        """

        p = list()
        x = self.endNode.pred
        while x.id is not 0:
            p.append(x.id)
            x = x.pred

        contractSet = p  # contracts serviced
        return self.endNode.val, contractSet  # profit and contracts serviced

    def __repr__(self):
        """
            Network description, can be expanded to give an initial desctiption
            of network.
        """
        return '(%s,%s,%s,%s,%s,%s)' % (str(self.nodes), str(self.arcs),
                                        str(self.startNode), str(self.endNode),
                                        str(self.solved), str(self.unsolved))
