import numpy as np
import pandas as pd

# Post-Processing


def color_leqzero_red(val):
    """
    Takes a scalar and returns a string with
    the css property `'color: red'` for negative
    strings, white otherwise.
    """
    color = 'red' if val <= 0 else 'white'
    return 'color: %s' % color


def highlight_max(data, color='blue'):
    '''
    highlight the maximum in a Series or DataFrame
    '''
    attr = 'background-color: {}'.format(color)
    if data.ndim == 1:  # Series from .apply(axis=0) or axis=1
        is_max = data == data.max()
        return [attr if v else '' for v in is_max]
    else:  # from .apply(axis=None)
        is_max = data == data.max().max()
        return pd.DataFrame(np.where(is_max, attr, ''),
                            index=data.index, columns=data.columns)


def getPostProcessing(T, rig_results, it_results, dict_rigs):
    """
        Main function for post-processing. Conduct local function calls
        for specific post-processing.

        Input:
        1. 3D array (S,P,R)
               S - number of contract scenarios, s - indicator (0...-1)
               P - number of properties, p
                   p = 0, profit
                   p = 1, number of contracts served
                   p = 2, total duration (months) of activity for contracts served
              R - number of rigs, r

        2. 2D array (S, P)
              S - number of contract scenarios, s - indicator (0...-1)
              P - number of properties, p
                   p = 0, number of contracts generated
                   p = 1, oil price
                   ... more properties will follow

        Output: 
        1. Dataframe with rig specific results - RxP array
            - All entries are average entires over S scenarios!
        2. Dataframe with scenario specific results - SxP' array
    """
    # NB: for means and sums, specify axis - 0=column i, all rows,
    #  1=row i all columns

    # Number of rigs
    numrigs = len(rig_results[0, 0])

    # ---- Create Rig Dataframe ---- #
    # ------------------------------ #

    # --- Profits --- #
    # Mean profit for each rig - 1xR array
    rig_mean_Revenue = np.mean(rig_results[:, 0], axis=0)
    rig_std_Revenue = np.std(rig_results[:, 0], axis=0)
    # Mean fleet profit - 1x1 scalar
    fleet_mean_Revenue = np.sum(rig_mean_Revenue)
    fleet_std_Revenue = np.sum(rig_std_Revenue)

    # --- Contracts Served --- #
    # Mean contracts for each rig - 1xR array
    rig_mean_ContractsServed = np.mean(rig_results[:, 1], axis=0)
    # Fleet mean contracts served - 1x1 scalar
    fleet_mean_ContractsServed = np.sum(rig_mean_ContractsServed)

    # --- Rig Utilisation --- #
    # Mean number of months on contract for each rig - 1xR array
    rig_mean_ActiveMonths = np.mean(
        rig_results[:, 2], axis=0)  # convert to years
    rig_std_ActiveMonths = np.std(
        rig_results[:, 2], axis=0)
    # Utilisation rate (active/idle) for each rig - 1xR array
    rig_mean_TotalMonths = np.mean(rig_results[:, 8], axis=0) + T*12.
    rig_rate_Utilisation = rig_mean_ActiveMonths / \
        (rig_mean_TotalMonths)  # sim. period + extra time if after 2030
    # Fleet mean utilisation - 1x1 scalar
    fleet_mean_ActiveMonths = np.sum(rig_mean_ActiveMonths)
    fleet_std_ActiveMonths = np.sum(rig_std_ActiveMonths)
    fleet_rate_Utilisation = np.mean(rig_rate_Utilisation)
    # fleet_rate_Utilisation = fleet_mean_ActiveMonths / (120.*8.) alternative

    # --- Rig Market Information --- #
    rig_mean_MarketTransitions = np.mean(rig_results[:, 3], axis=0)
    rig_mean_HEcontracts = np.mean(rig_results[:, 4], axis=0)
    rig_mean_ExplApprContracts = np.mean(rig_results[:, 5], axis=0)
    rig_mean_MidwaterContracts = np.mean(rig_results[:, 6], axis=0)
    rig_mean_DeepwaterContracts = np.mean(rig_results[:, 7], axis=0)

    fleet_mean_MarketTransitions = np.sum(rig_mean_MarketTransitions)
    fleet_mean_HEcontracts = np.sum(rig_mean_HEcontracts)
    fleet_mean_ExplApprContracts = np.sum(rig_mean_ExplApprContracts)
    fleet_mean_MidwaterContracts = np.sum(rig_mean_MidwaterContracts)
    fleet_mean_DeepwaterContracts = np.sum(rig_mean_DeepwaterContracts)

    # --- Costs --- #
    # NB: should be constant
    rig_mean_Opex = np.mean(rig_results[:, 9], axis=0)
    rig_mean_Capex = np.mean(rig_results[:, 10], axis=0)
    fleet_mean_Opex = np.sum(rig_mean_Opex)
    fleet_mean_Capex = np.sum(rig_mean_Capex)

    # --------------------------- #

    # Add entries to rig dictionary - all entries are average entires over S scenarios
    dict_rigs.update({'Mean revenue (MUSD)': rig_mean_Revenue,
                      'Stddev revenue (MUSD)': rig_std_Revenue,
                      'Contracts served': rig_mean_ContractsServed,
                      'Mean active years': rig_mean_ActiveMonths/12.,  # no of years on contract
                      'Stddev active years': rig_std_ActiveMonths/12.,
                      'Utilisation': rig_rate_Utilisation,  # rate of active over total time
                      'OPEX': rig_mean_Opex,
                      'CAPEX': rig_mean_Capex,
                      'Market transitions': rig_mean_MarketTransitions,
                      'HE contracts': rig_mean_HEcontracts,
                      'Exploration contracts': rig_mean_ExplApprContracts,
                      '150-1500m contracts': rig_mean_MidwaterContracts,
                      '>1500m contracts': rig_mean_DeepwaterContracts,
                      })
    # Create rig dataframe
    rig_df = pd.DataFrame(dict_rigs, columns=['Name', 'HE/BE', 'Water depth(m)',
                                              'Mean revenue (MUSD)', 'Stddev revenue (MUSD)',
                                              'Contracts served', 'Mean active years',
                                              'Stddev active years', 'Utilisation',
                                              'OPEX', 'CAPEX',
                                              'Market transitions', 'HE contracts',
                                              'Exploration contracts', '150-1500m contracts',
                                              '>1500m contracts'])
    # Create fleet dataframe - all values are averages
    rig_df_tot = pd.DataFrame({'Name': ['Total Fleet'],
                               'HE/BE': ['-'],
                               'Water depth(m)': ['-'],
                               'Mean revenue (MUSD)': fleet_mean_Revenue,
                               'Stddev revenue (MUSD)': fleet_std_Revenue,
                               'Contracts served': fleet_mean_ContractsServed,
                               'Mean active years': fleet_mean_ActiveMonths/12.,
                               'Stddev active years': fleet_std_ActiveMonths/12.,
                               'Utilisation': fleet_rate_Utilisation,
                               'OPEX': fleet_mean_Opex,
                               'CAPEX': fleet_mean_Capex,
                               'Market transitions': fleet_mean_MarketTransitions,
                               'HE contracts': fleet_mean_HEcontracts,
                               'Exploration contracts': fleet_mean_ExplApprContracts,
                               '150-1500m contracts': fleet_mean_MidwaterContracts,
                               '>1500m contracts': fleet_mean_DeepwaterContracts,
                               })
    # Append fleet values to rig dataframe
    rig_df = rig_df.append(rig_df_tot, ignore_index=True, sort=False)

    # Specify number of decimals
    rig_df = rig_df.round(2)

    # Set rig name equal to index
    # rig_df.set_index('Name', inplace=True)

    # ---- Styling ---- # play around
    # applymap works elementwise
    # apply works serieswise
    # NB: use method chains to build style

    # rig_df = rig_df.style.\
    #     format({'Profit (MUSD)': lambda x: "{:.2f}".format(round(x, 2)),
    #             'Contracts served': lambda x: "{:.2f}".format(round(x, 2)),
    #             'Active years': lambda x: "{:.2f}".format(round(x, 2)),
    #             'Utilisation': "{:.2%}"})  # . \

    # applymap(color_leqzero_red, subset=[
    #     'Profit (MUSD)', 'Contracts served', 'Active years', 'Utilisation']).\
    # apply(highlight_max, axis=0, subset=pd.IndexSlice[0:numrigs-1, [
    #       'Profit (MUSD)', 'Contracts served', 'Active years', 'Utilisation']])

    # axis=None, tablewise style
    # axis=0, columnwise style
    # axis=1, rowwise style

    # subset specifies columns to apply function to => slicing

    # ---- Create Iteration Dataframe ---- #
    # ----------------------------------- #

    # ---- Contract Service ---- #
    # Total contracts served for each scenario - 1xS array
    it_total_ContractsServed = [np.sum(rig_results[i, 1], axis=0)
                                for i in range(len(rig_results))]
    # Rate of contracts served vs contracts generated - 1xS array
    it_rate_ContractService = it_total_ContractsServed / it_results[:, 0]

    # ---- Environment ---- #
    # Rate of HE contracts

    # ---- Water depth ---- #
    # Rate of deepwater contracts

    # ---- Contract Revenue ---- #
    # Average dayrate

    # --------------------------- #
    # Create contract dictionary
    dict_iterations = {'Iteration': [i+1 for i in range(len(it_results[:, 0]))],
                       'Contracts generated': it_results[:, 0],
                       'Served contracts': it_total_ContractsServed,
                       'Service rate': it_rate_ContractService,
                       'Revenue (MUSD)': it_results[:, 1],
                       'Cost (MUSD)': it_results[:, 3],
                       'Utilisation': it_results[:, 2],
                       'Oil price': it_results[:, 4], }

    # Create contract dataframe
    it_df = pd.DataFrame(dict_iterations, columns=[
        'Iteration', 'Contracts generated', 'Served contracts', 'Service rate',
        'Revenue (MUSD)', 'Cost (MUSD)', 'Utilisation', 'Oil price', ])

    it_df = it_df.round(2)
    # Set scenario number as index
    # it_df.set_index('Scenario', inplace=True)

    return rig_df, it_df


if __name__ == '__main__':

    rig_df, it_df = getPostProcessing(rig_results, it_results, dict_rigs)
